"""
    Convenience methods for using the library.

    Example:
        >>> import ipdata
        >>> ipdata.api_key = <YOUR API KEY>
        >>> ipdata.lookup() # or ipdata.lookup("8.8.8.8") 
"""
from .ipdata import IPData

# Configuration
api_key = None
endpoint = None
default_client = None


def lookup(resource, fields=[]):
    return _proxy("lookup", resource=resource, fields=fields)


def bulk(resources, fields=[]):
    return _proxy("bulk", resources, fields=fields)


def _proxy(method, *args, **kwargs):
    """Create an IPData client if one doesn't exist."""
    global default_client
    if not default_client:
        default_client = IPData(
            api_key,
        )

    fn = getattr(default_client, method)
    return fn(*args, **kwargs)



import ipaddress
import os

import pytest
from dotenv import load_dotenv
from hypothesis import given, settings
from hypothesis import strategies as st

from pathlib import Path
import ipdata

# local testing
pwd = Path(__file__).parent.resolve()
load_dotenv(f"{pwd}/.env")

# Github CI runs
IPDATA_API_KEY = os.environ.get("IPDATA_API_KEY")
ipdata.api_key = IPDATA_API_KEY


@given(st.ip_addresses(network="8.8.8.0/24"))
@settings(deadline=None, max_examples=100)
@pytest.mark.api
def test_lookup_v4(ip):
    ip = str(ip)
    data = ipdata.lookup(ip)
    assert data.ip == ip


@given(st.ip_addresses(network="2620:11a:a000::/40"))
@settings(deadline=None, max_examples=100)
@pytest.mark.api
def test_lookup_v6(ip):
    ip = str(ip)
    data = ipdata.lookup(ip)

    # Capitalize to handle 2620:11A:A000::' == '2620:11a:a000::'
    assert data.ip.upper() == ip.upper()


@given(
    st.lists(st.ip_addresses(network="2620:11a:a000::/40"), min_size=1, max_size=100)
)
@settings(deadline=None, max_examples=100)
@pytest.mark.api
def test_bulk_v6(ips):
    ips = [str(ip) for ip in ips]
    data = ipdata.bulk(ips)

    # Check we got all results
    assert len(data.responses) == len(ips)

    # Capitalize to handle 2620:11A:A000::' == '2620:11a:a000::'
    for pair in zip(ips, [response.ip for response in data.responses]):
        assert pair[0].upper() == pair[0].upper()


@given(st.lists(st.ip_addresses(network="8.8.8.0/24"), min_size=1, max_size=100))
@settings(deadline=None, max_examples=100)
@pytest.mark.api
def test_bulk_v4(ips):
    ips = [str(ip) for ip in ips]
    data = ipdata.bulk(ips)

    # Check we got all results
    assert len(data.responses) == len(ips)

    for pair in zip(ips, [response.ip for response in data.responses]):
        assert pair[0] == pair[0]



"""This is the official IPData client library for Python.

IPData client libraries allow for looking up the geolocation,
ownership and threat profile or any IP address and ASN.

Example:
    >>> from ipdata import IPData
    >>> ipdata = IPData("YOUR API KEY")
    >>> ipdata.lookup() # or ipdata.lookup("8.8.8.8")

Alternatively thanks to the convenience methods in __init__.py you can do

Example
    >>> import ipdata
    >>> ipdata.api_key = <YOUR API KEY>
    >>> ipdata.lookup() # or ipdata.lookup("8.8.8.8") 

:class:`~.IPData` is the primary class for making API requests.
"""

import ipaddress
import requests
import logging
import urllib3

from requests.adapters import HTTPAdapter, Retry
from rich.logging import RichHandler

FORMAT = "%(message)s"
logging.basicConfig(
    level="ERROR", format=FORMAT, datefmt="[%X]", handlers=[RichHandler()]
)


class IPDataException(Exception):
    pass


class DotDict(dict):
    """
    dot.notation access to dictionary attributes
    Based on https://stackoverflow.com/a/23689767/3176550
    """

    def __getattr__(*args):
        val = dict.get(*args)
        if type(val) is dict:
            return DotDict(val)
        elif type(val) is list:
            return [DotDict(item) for item in val]
        return val

    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__


class IPData(object):
    """
    Instances of IPData are used to make API requests. To make requests to the EU endpoint, set "endpoint" to "https://eu-api.ipdata.co".

    :param api_key: A valid IPData API key
    :param endpoint: The API endpoint to send requests to
    :param timeout: The default requests timeout
    :param retry_limit: The maximum number of retries in case of failure
    :param retry_backoff_factor: The number of seconds to sleep in between retries. With a 1 second backoff calls with be retried up to 7 times at the following intervals: 0.5, 1, 2, 4, 8, 16, 32
    :param debug: A boolean used to set the log level. Set to True when debugging.
    """

    log = logging.getLogger("rich")

    valid_fields = {
        "ip",
        "is_eu",
        "city",
        "region",
        "region_code",
        "country_name",
        "country_code",
        "continent_name",
        "continent_code",
        "latitude",
        "longitude",
        "asn",
        "postal",
        "calling_code",
        "flag",
        "emoji_flag",
        "emoji_unicode",
        "carrier",
        "languages",
        "currency",
        "time_zone",
        "threat",
        "count",
        "status",
        "company",
    }

    def __init__(
        self,
        api_key,
        endpoint="https://api.ipdata.co/",
        timeout=60,
        retry_limit=7,
        retry_backoff_factor=1,
        debug=False,
    ):
        # Request settings
        self.api_key = api_key
        self.endpoint = endpoint.rstrip("/")  # remove trailing /
        self._timeout = timeout
        self._headers = {"user-agent": "ipdata-python"}  # set default UserAgent
        self._query_params = {"api-key": self.api_key}

        # Enable debugging
        if debug:
            self.log.setLevel(logging.DEBUG)

        # Work around renamed argument in urllib3.
        if hasattr(urllib3.util.Retry.DEFAULT, "allowed_methods"):
            methods_arg = "allowed_methods"
        else:
            methods_arg = "method_whitelist"

        # Retry settings
        retry_args = {
            "total": retry_limit,
            "backoff_factor": retry_backoff_factor,
            "status_forcelist": [429, 500, 502, 503, 504],
            methods_arg: {"POST"},
        }

        adapter = HTTPAdapter(max_retries=urllib3.Retry(**retry_args))

        self._session = requests.Session()
        self._session.mount("http", adapter)

    def _validate_fields(self, fields):
        """
        Validates the fields passed in by the user, first ensuring it's a collection. In prior versions 'fields' was a string, however it now needs to be a collection.

        :param fields: A collection of supported fields
        :raises ValueError: if 'fields' is not one of ['list', 'tuple' or 'set'] or if 'fields' contains unsupported fields
        """
        if not type(fields) in [list, set, tuple]:
            raise ValueError("'fields' must be of type 'list', 'tuple' or 'set'.")

        # Get all unsupported fields
        diff = set(fields).difference(self.valid_fields)
        if diff:
            raise ValueError(
                f"The field(s) {diff} are not supported. Only {self.valid_fields} are supported."
            )

    def _validate_ip_address(self, ip):
        """
        Checks that 'ip' is a valid IP Address.

        :param ip: A string
        :raises ValueError: if 'ip' is either not a valid IP address or is a reserved IP address eg. private, reserved or multicast
        """
        request_ip = ipaddress.ip_address(ip)
        if request_ip.is_private or request_ip.is_reserved or request_ip.is_multicast:
            raise ValueError(f"{ip} is a reserved IP Address")

    def lookup(self, resource="", fields=[]):
        """
        Makes a GET request to the IPData API for the specified 'resource' and the given 'fields'.

        :param resource: Either an IP address or an ASN prefixed by "AS" eg. "AS15169"
        :param fields: A collection of API fields to be returned

        :returns: An API response as a DotDict object to allow dot notation access of fields eg. data.ip, data.company.name, data.threat.blocklists[0].name etc

        :raises IPDataException: if the API call fails or if there is a failure in decoding the response.
        :raises ValueError: if 'resource' is not a string
        """
        if type(resource) is not str:
            raise ValueError(f"{resource} must be of type 'str'")

        resource = (
            resource.upper()
        )  # capitalize resource in case it's a typoed ASN query eg. as2

        if resource and not resource.startswith("AS"):
            self._validate_ip_address(resource)

        self._validate_fields(fields)
        query_params = self._query_params | {"fields": ",".join(fields)}

        # Make the request
        try:
            response = self._session.get(
                f"{self.endpoint}/{resource}",
                headers=self._headers,
                params=query_params,
                timeout=self._timeout,
            )
        except Exception as e:
            raise IPDataException(e)

        status_code = response.status_code

        # Decode the response and add metadata
        try:
            data = DotDict(response.json())
            data["status"] = status_code
        except ValueError:
            raise IPDataException(
                f"An error occured while decoding the API response: {response.text}"
            )

        return data

    def bulk(self, resources, fields=[]):
        """
        Lookup up to 100 resources in one request. Makes a POST request wth the resources as a JSON array and the specified fields.

        :param resources: A list of resources. This can be a list of IP addresses or ASNs or a mix of both which is not recommended unless you handle it during parsing. Mixing resources might lead to weird behavior when writing results to CSV.
        :param fields: A collection of API fields to be returned.

        :returns: A DotDict object with the data contained under the 'responses' key.

        :raises ValueError: if resources it not a collection
        :raises ValueError: if resources contains 0 items
        :raises IPDataException: if the API call fails or if there is an error decoding the response
        """

        if type(resources) not in [list, set]:
            raise ValueError(f"{resources} must be of type 'list' or 'set'")

        if len(resources) < 1:
            raise ValueError("Bulk lookups must contain at least 1 resource.")

        self._validate_fields(fields)
        query_params = self._query_params | {"fields": ",".join(fields)}

        # Make the requests
        try:
            response = self._session.post(
                f"{self.endpoint}/bulk",
                headers=self._headers,
                params=query_params,
                json=resources,
            )
        except Exception as e:
            raise IPDataException(f"Error when looking up {resources} - {e}")

        status_code = response.status_code

        # Decode the response
        try:
            data = response.json()
        except ValueError:
            raise IPDataException(
                f"An error occured while decoding the API response: {response.text}"
            )

        # If the request returned a non 200 response we add metadata and return the response
        if not status_code == 200:
            data["status"] = status_code
            return data

        return DotDict(
            {
                "responses": [DotDict(resource) for resource in data],
                "status": status_code,
            }
        )



"""
    Geofeeds classes used in the CLI for the validator.
"""
import csv
import hashlib
import ipaddress
import logging
import random
from pathlib import Path

import requests
from rich.logging import RichHandler

from .codes import COUNTRIES, REGION_CODES

FORMAT = "%(message)s"
logging.basicConfig(
    level="ERROR", format=FORMAT, datefmt="[%X]", handlers=[RichHandler()]
)
log = logging.getLogger("rich")

pwd = Path(__file__).parent.resolve()


class GeofeedValidationError(Exception):
    pass


class Geofeed(object):
    """
    Create an instance of a Geofeed object.

    :param source: Either a URL or a local file containing geofeed formatted data according to https://datatracker.ietf.org/doc/html/draft-google-self-published-geofeeds-05.
    :param dir: The directory where the downloaded copy of a geofeed is cached
    """

    def __init__(self, source, dir="/tmp") -> None:
        self.source = source
        self.dir = dir
        self.cache_path = None
        self.total_count = 0
        self.valid_count = 0
        # Ensure uniqueness https://datatracker.ietf.org/doc/html/draft-google-self-published-geofeeds-05#appendix-A
        self.prefixes = set()

    def _download(self):
        """
        Downloads and caches the geofeed if a valid URL is provided

        :raises Exception: if a failure occurs when downloading the geofeed
        """
        random_name = hashlib.sha224( str(random.getrandbits(256)).encode('utf-8') ).hexdigest()[:16]
        cache_path = f"{pwd}/{self.dir}/{random_name}.csv"

        try:
            response = requests.get(self.source, timeout=60)
            text = response.text
            with open(cache_path, "w") as f:
                f.write(
                    "\n".join(
                        line for line in text.split("\n") if not line.startswith("#")
                    )
                )  # don't write comments
        except Exception:
            log.exception(f"Failed to get {self.source}")
        else:
            # set cache path if download was successful
            self.cache_path = cache_path

    def entries(self):
        """
        Reads each line in a geofeed and yields Entry objects, one for each line.

        :raises GeofeedValidationError: if it find a duplicate or if the expected CSV format is invalid eg. if the line has != 5 columns. There are a number of optional fields however the minimum number of commans should be present i.e. 4.
        Also note that though the RFC recommends simply discarding extra columns to allow for additional fields in the future, we strictly check for 5 columns to prevent breaking out CSV parsing.
        :yields: An Entry object
        """
        # set path to the source by default
        path = self.source

        # if the source is a url and we haven't already download it, try to download it and update path to the location of the download cache
        if "://" in self.source and not self.cache_path:
            self._download()
            path = self.cache_path

        # if path still contains a url it means the download was not successful in which case we yield nothing!
        if "://" in path:
            log.warning(f"No entries found {self.source}")
            yield

        with open(path) as f:
            reader = csv.reader(f)
            for entry in reader:
                # keep count of the number of entries
                self.total_count += 1

                # generate error if there are not exactly 5 fields
                if len(entry) != 5:
                    yield GeofeedValidationError(
                        f"[{entry}] is not a valid geofeed entry. The 'IP Range' and 'Country' fields are required however you must have the requisite minimum number of commas (currently 4) present"
                    )

                # instantiate an Entry object
                geofeed_entry = Entry(*entry)

                # check for duplicates and generate error if found
                if geofeed_entry.ip_range in self.prefixes:
                    yield GeofeedValidationError(
                        f"Duplicate prefixes found {geofeed_entry.ip_range}"
                    )

                self.prefixes.add(geofeed_entry.ip_range)

                # generate entry
                yield geofeed_entry


class Entry(object):
    """
    Create an instance of an Entry object

    :param ip_range: A valid IP address prefix
    :param country: A valid ISO 3166-1 alpha 2 country code
    :param region: A valid ISO 3166-1 alpha 2 subdivision code
    :param city: The city where the IP range is allocated
    :param postal_code: The postal code of the location the IP range is allocated
    """

    def __init__(
        self, ip_range, country, region=None, city=None, postal_code=None
    ) -> None:
        self.ip_range = ip_range
        self.country = country.upper()
        self.region = region
        self.city = city
        self.postal_code = postal_code
        self.row = [
            self.ip_range,
            self.country,
            self.region,
            self.city,
            self.postal_code,
        ]

    def __str__(self) -> str:
        return f"{self.ip_range},{self.country},{self.region},{self.city},{self.postal_code}"

    def validate(self):
        """
        Validation rules for geofeed entries. Note that though the spec states that all fields other than the IP range are optional, we require at the minimum at country code. Entries without a country code will not pass validation.

        :raises GeofeedValidationError: if any of the validation rules are violated
        """

        # 1. Each IP range field MUST be either a single IP address or an IP prefix in CIDR notation in conformance with section 3.1 of[RFC4632] for IPv4 or section 2.3 of [RFC4291] for IPv6.
        # Reference: https://datatracker.ietf.org/doc/html/draft-google-self-published-geofeeds-02#section-2.1.1.1

        try:
            ipaddress.ip_network(self.ip_range)
        except ValueError:
            raise GeofeedValidationError(
                f"[{self}] does not have a valid IP address or prefix"
            )

        # 2 (a) Country code must be present at a minimum

        if not self.country:
            raise GeofeedValidationError(f"[{self}] is missing a country code")

        # 2 (b) Country code must be a valid 2-letter ISO 3166-1 alpha 2 country code
        # Reference: https://datatracker.ietf.org/doc/html/draft-google-self-published-geofeeds-02#section-2.1.1.2
        if not self.country in COUNTRIES:
            raise GeofeedValidationError(
                f"[{self}] ({self.country}) is not a valid ISO 3166-1 alpha 2 subdivision code"
            )

        # The region code is optional, but if it exists we need to validate it
        if self.region:
            # 3 (a) The region field, if non-empty, MUST be a ISO region code conforming to ISO 3166-2 [ISO.3166.2].  Parsers SHOULD treat this field case-insensitively.
            # Reference: https://datatracker.ietf.org/doc/html/draft-google-self-published-geofeeds-02#section-2.1.1.3
            if not self.region in REGION_CODES:
                raise GeofeedValidationError(
                    f"[{self}] ({self.region}) is not a valid ISO 3166-1 alpha 2 region code"
                )

            # 3 (b) The region must be in the same country
            if not self.region.split("-")[0] == self.country:
                raise GeofeedValidationError(
                    f"[{self}] the region code provided is in a different country"
                )



"""This is the official IPData Command Line Interface.

Use it to do one-off lookups or high throughput bulk lookups. With a ton of convenience features eg.
copying a result to the clipboard with '-c', pretty printing results in easy to parse panels with '-p' and more!

    $ ipdata --help
    Usage: ipdata [OPTIONS] COMMAND [ARGS]...

      Welcome to the ipdata CLI

    Options:
      --api-key TEXT  Your ipdata API Key. Get one for free from
                      https://ipdata.co/sign-up.html
      --help          Show this message and exit.

    Commands:
      lookup*
      batch
      init
      usage
      validate   

    $ ipdata
    Your IP

    $ ipdata 1.1.1.1 -f ip -f country_name
    {
      "ip": "1.1.1.1",
      "asn": {
        "asn": "AS13335",
        "name": "Cloudflare, Inc.",
        "domain": "cloudflare.com",
        "route": "1.1.1.0/24",
        "type": "business"
      },
      "status": 200
    }

    $ ipdata 1.1.1.1 -f ip -f asn -c 
    📋️ Copied result to clipboard!  

    $ ipdata 8.8.8.8 -f ip -p
    ╭───────────────╮
    │ country_name  │
    │ United States │
    ╰───────────────╯ 
"""
import csv
import io
import json
import logging
import os
import sys
from concurrent.futures import ThreadPoolExecutor, as_completed
from pathlib import Path
from sys import stderr

import click
import pyperclip
from click_default_group import DefaultGroup
from rich import print, print_json
from rich.columns import Columns
from rich.console import Console
from rich.logging import RichHandler
from rich.panel import Panel
from rich.progress import Progress
from rich.tree import Tree

from .lolcat import LolCat
from .geofeeds import Geofeed, GeofeedValidationError
from .ipdata import DotDict, IPData

console = Console()

FORMAT = "%(message)s"
logging.basicConfig(
    level="ERROR", format=FORMAT, datefmt="[%X]", handlers=[RichHandler()]
)
log = logging.getLogger("rich")

API_KEY_FILE = f"{Path.home()}/.ipdata"


def _lookup(ipdata, *args, **kwargs):
    """
    Wrapper for looking up individual resources with error handling. Takes the same arguments and returns the same response as ipdata.lookup.
    """
    try:
        response = ipdata.lookup(*args, **kwargs)
    except Exception as e:
        log.error(e)
    else:
        return response


def print_ascii_logo():
    """
    Print cool ascii logo with lolcat.
    """
    options = DotDict({"animate": False, "os": 6, "spread": 3.0, "freq": 0.1})
    logo = """
 _           _       _        
(_)_ __   __| | __ _| |_ __ _ 
| | '_ \ / _` |/ _` | __/ _` |
| | |_) | (_| | (_| | || (_| |
|_| .__/ \__,_|\__,_|\__\__,_|
  |_|
    """
    lol = LolCat()
    lol.cat(io.StringIO(logo), options)


def pretty_print_data(data):
    """
    Users rich to generate panels for individual API response fields for better readability!

    :param data: the response from ipdata.lookup
    """
    # we print single value panels first then multiple value panels for better organization
    single_value_panels = []
    multiple_value_panels = []

    # if data is empty do nothing
    if not data:
        return

    # push the blocklists field up a level, it's usually nested under the threat data
    if data.get("threat", {}).get("blocklists"):
        data["blocklists"] = data.get("threat", {}).pop("blocklists")

    # generate panels!
    for key, value in data.items():
        # simple case
        if type(value) in [str, bool]:
            single_value_panels.append(
                Panel(f"[b]{key}[/b]\n[yellow]{value}", expand=True)
            )

        # if the value is a dictionary we generate a tree inside a panel
        if type(value) is dict:
            tree = Tree(key)
            for k, v in value.items():
                if key == "threat":
                    if v:
                        sub_tree = tree.add(f"[b]{k}[/b]\n[bright_red]{v}")
                    else:
                        sub_tree = tree.add(f"[b]{k}[/b]\n[green]{v}")
                else:
                    sub_tree = tree.add(f"[b]{k}[/b]\n[yellow]{v}")
            multiple_value_panels.append(Panel(tree, expand=False))

        # if value if a list we generate nested trees
        if type(value) is list:
            tree = Tree(key)
            for item in value:
                branch = tree.add("")
                for k, v in item.items():
                    _ = branch.add(f"[b]{k}[/b]\n[yellow]{v}")
            multiple_value_panels.append(Panel(tree, expand=False))

    # print the single value panels to the console
    console.print(Columns(single_value_panels), overflow="ignore", crop=False)

    # print the multiple value panels to the console
    console.print(Columns(multiple_value_panels), overflow="ignore", crop=False)


@click.group(
    help="Welcome to the ipdata CLI",
    cls=DefaultGroup,
    default_if_no_args=True,
    default="lookup",
)
@click.option(
    "--api-key",
    required=False,
    default=None,
    help="Your ipdata API Key. Get one for free from https://ipdata.co/sign-up.html",
)
@click.pass_context
def cli(ctx, api_key):
    """
    This is the main entry point for the CLI. We first check for the presence of an API key at ~/.ipdata and make it accessible to all other commands in our group.
    Note that the 'init' and 'validate' commands are the only ones that can run without an API key present.

    :param api_key: A valid IPData API key
    """
    ctx.ensure_object(dict)
    key = ctx.obj["api-key"] = (
        Path(API_KEY_FILE).read_text() if Path(API_KEY_FILE).exists() else None
    )

    # Allow init and validate to run without an API key
    if not ctx.invoked_subcommand in ["init", "validate"] and key is None:
        print(
            f'Please initialize the cli by running "ipdata init <api key>" then try again',
            file=stderr,
        )
        sys.exit(1)


@cli.command()
@click.argument(
    "api-key",
    required=True,
    type=str,
)
def init(api_key):
    """
    Initialize the CLI by setting an API key.

    :param api_key: A valid IPData API key
    """
    ipdata = IPData(api_key)

    # We verify that the API key can successfully make requests by making one
    with console.status("Verifying key ...", spinner="dots12"):
        response = _lookup(ipdata, "8.8.8.8")

    # Handle success
    if response.status == 200:
        with open(API_KEY_FILE, "w") as f:
            f.write(api_key)
        print_ascii_logo()
        print(f"✨ Successfully initialized.")
    else:
        # Handle failure
        print(
            f"Initialization failed. Error: {response.status}): {response.message}",
            file=stderr,
        )


@cli.command()
@click.option("--api-key", "-k", required=False, default=None, help="ipdata API Key")
@click.pass_context
def usage(ctx, api_key):
    """
    Get today's usage. Quota resets every day at 00:00 UTC.
    """
    api_key = ctx.obj["api-key"]
    ipdata = IPData(api_key)
    with console.status("Getting usage ...", spinner="dots12"):
        response = _lookup(ipdata)
    print(f"{int(response.count):,}")


@cli.command(default=True)
@click.argument("resource", required=False, type=str, default="")
@click.option(
    "--fields", "-f", required=False, multiple=True, default=[],
)
@click.option(
    "--exclude", "-e", required=False, multiple=True, default=[],
)
@click.option("--api-key", "-k", required=False, default=None, help="ipdata API Key")
@click.option(
    "--pretty-print",
    "-p",
    is_flag=True,
    required=False,
    default=False,
    help="Pretty prints results as panels",
)
@click.option(
    "--raw",
    "-r",
    is_flag=True,
    required=False,
    default=False,
    help="Disable pretty printing",
)
@click.option(
    "--copy",
    "-c",
    is_flag=True,
    required=False,
    default=False,
    help="Copy the result to the clipboard",
)
@click.pass_context
def lookup(ctx, resource, fields, api_key, pretty_print, raw, copy, exclude):
    """
    Lookup resources by using the IPData class methods.

    :param resource: The resource to lookup
    :param fields: A list of supported fields passed as multiple parameters eg. "... -f ip -f country_name"
    :param api_key: A valid API key
    :param pretty_print: Whether to pretty print the response with panels
    :param raw: Whether to print raw unformatted but syntax-highlighted JSON
    :param copy: Copy the response to the clipboard
    """
    api_key = api_key if api_key else ctx.obj["api-key"]
    ipdata = IPData(api_key)

    # enforce mutual exclusivity of fields and exclude
    if exclude and fields:
        raise click.ClickException("'--fields / -f' and '--exclude / -e' are mutually exclusive.")

    # if the user wants to exclude some fields, get all the fields in fields that are not in exclude
    if exclude:
        fields = set(ipdata.valid_fields).difference(set(exclude))

    with console.status(
        f"""Looking up {resource if resource else "this device's IP address"}""",
        spinner="dots12",
    ):
        data = _lookup(ipdata, resource, fields=fields)

    if copy:
        pyperclip.copy(json.dumps(data))
        print(f"📋️ Copied result to clipboard!")
    elif raw:
        print(data)
    elif pretty_print:
        pretty_print_data(data)
    else:
        print_json(data=data)


def chunks(lst, n=100):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i : i + n]


def process(resources, processor, fields):
    """Farm out work to worker threads."""
    n_workers = os.cpu_count()
    with ThreadPoolExecutor(n_workers) as executor:
        futures = [
            executor.submit(processor, chunk, fields) for chunk in chunks(resources)
        ]
        for future in as_completed(futures):
            try:
                result = future.result()
            except Exception as e:
                log.error(e)
            else:
                yield result


@cli.command()
@click.argument("input", required=True, type=click.File(mode="r", encoding="utf-8"))
@click.option(
    "--fields", "-f", required=False, multiple=True
)
@click.option(
    "--exclude", "-e", required=False, multiple=True, default=[],
)
@click.option(
    "--output", "-o", required=True, type=click.File(mode="w", encoding="utf-8")
)
@click.option(
    "--format",
    required=False,
    type=click.Choice(("JSON", "CSV"), case_sensitive=False),
    default="JSON",
    help="Output file format",
)
@click.pass_context
def batch(ctx, input, fields, output, format, exclude):
    """
    Batch command for doing fast bulk processing.

    :param input: A list of resources to lookup either IP addresses or ASNs. You could mix them however this could break output processing when writing to CSV
    :param fields: A list of valid fields to extract from the individual responses
    :param output: The path to write the results to
    :param format: The format to write the results to. When using the CSV format it is required to provide fields.
    """

    if format == "CSV" and not fields:
        print(
            f'You need to specify a "--fields" argument with a list of fields to extract to get results in CSV. To get entire responses use JSON.',
            file=stderr,
        )
        return

    # enforce mutual exclusivity of fields and exclude
    if exclude and fields:
        raise click.ClickException("'--fields / -f' and '--exclude / -e' are mutually exclusive.")

    # if the user wants to exclude some fields, get all the fields in fields that are not in exclude
    if exclude:
        fields = set(ipdata.valid_fields).difference(set(exclude))


    # Prepare requests
    ipdata = IPData(ctx.obj["api-key"])
    resources = [resource.strip() for resource in input.readlines()]
    bulk_results = process(resources, ipdata.bulk, fields)

    # Prepare CSV writing by expanding fieldnames eg. asn to asn, name, domain etc
    csv_writer = None
    fieldnames = []
    for field in fields:
        if field == "asn":
            fieldnames += [
                f"asn_{sub_field}"
                for sub_field in ["asn", "name", "domain", "route", "type"]
            ]
            continue
        if field == "company":
            fieldnames += [
                f"company_{sub_field}"
                for sub_field in ["asn", "name", "domain", "network", "type"]
            ]
            continue
        if field == "threat":
            fieldnames += [
                f"asn_{sub_field}"
                for sub_field in [
                    "is_tor",
                    "is_icloud_relay",
                    "is_proxy",
                    "is_datacenter",
                    "is_anonymous",
                    "is_known_attacker",
                    "is_known_abuser",
                    "is_threat",
                    "is_bogon",
                    "blocklists",
                ]
            ]
            continue
        if field in ipdata.valid_fields:
            fieldnames += [field]

    # Do lookups concurrenctly using threads in batches of 100 each
    with Progress() as progress:
        # update progress bar
        task = progress.add_task("[green]Processing...", total=len(resources))

        # write individual results to file
        for bulk_result in bulk_results:
            for result in bulk_result.get("responses", {}):
                progress.update(task, advance=1)
                if format == "JSON":
                    output.write(f"{result}\n")
                if format == "CSV":
                    if not csv_writer:
                        # create writer if none exists
                        csv_writer = csv.DictWriter(output, fieldnames=fieldnames)
                        csv_writer.writeheader()

                    # prepare row
                    row = {}
                    for field, value in result.items():
                        if type(value) is dict:
                            for k, v in result[field].items():
                                row[f"{field}_{k}"] = v
                        else:
                            row[field] = value

                    # ensure no unexpected fields
                    try:
                        # handle when dict fields are none eg. when asn, company or carrier are empty
                        row_copy = row.copy()
                        for key in row_copy:
                            if key not in fieldnames:
                                row.pop(key)

                        # write results
                        csv_writer.writerow(row)
                    except ValueError as e:
                        log.error(f"Error writing row: {row}. Error: {e}")


@cli.command()
@click.argument("feed", required=True, type=str)
def validate(feed):
    """
    Validates a geofeed file.

    :param feed: Either a valid URL (with the https:// path) or a file path
    """
    geofeed = Geofeed(feed)
    valid = True
    for entry in geofeed.entries():
        if type(entry) is GeofeedValidationError:
            log.error(entry)
            valid = False
        else:
            try:
                entry.validate()

                # keep count of valid entries
                geofeed.valid_count += 1
            except GeofeedValidationError as e:
                valid = False
                log.error(e)

    if geofeed.total_count == 0:
        log.error(f"The provided geofeed is empty")

    valid_percentage = geofeed.valid_count / geofeed.total_count * 100
    print(
        f"{geofeed.source} has {geofeed.valid_count:,} ({valid_percentage:.2f}%) valid entries."
    )

    if not valid:
        sys.exit(1)

    print(
        "✨ Success! Your geofeed is valid and ready for publishing! Send an email to corrections@ipdata.co with the URL of this feed."
    )


if __name__ == "__main__":
    cli()



import pytest

from .geofeeds import Entry, GeofeedValidationError


@pytest.mark.parametrize(
    "entry",
    [
        "172.32.0.0/55,,,,",
        "172.32.0.0/11,,,,",
        "172.32.0.0/11,ZZ,,,",
        "172.32.0.0/11,US,ZZ-ZZ,,",
        "172.32.0.0/11,US,LU-LU,,",
    ],
)
@pytest.mark.geofeed
def test_validation_rules(entry):
    entry = Entry(*entry.split(","))
    with pytest.raises(GeofeedValidationError):
        entry.validate()
